# -*- coding: utf-8 -*-

from __future__ import absolute_import

import os
import sys

import pytest

try:
    from shlex import quote
except ImportError:
    from pipes import quote  # noqa


@pytest.fixture()
def delegator():
    """Execute a command and return its value."""
    import delegator

    def run(command):
        if "|" in command:
            res = delegator.chain(command)
        else:
            res = delegator.run(command)
        if res.return_code == 127:
            raise FileNotFoundError(res.err)
        if res.return_code != 0:
            raise RuntimeError(res.err)
        if res.err:
            sys.stderr.write(res.err)
            if not res.err.endswith(os.linesep):
                sys.stderr.write(os.linesep)

        return res.out.strip()

    return run


@pytest.fixture()
def git_command(delegator, git_directory):
    """A delegator to ``git`` setting the appropriate ``--git-dir``."""

    def run(cmd, **kwargs):
        cmd = "git --git-dir={} {}".format(quote(git_directory), cmd)
        return delegator(cmd, **kwargs)

    return run


@pytest.fixture()
def git_directory(delegator):
    """The ``.git`` directory for the current project."""
    return delegator("git rev-parse --git-dir")



# -*- coding: utf-8 -*-

from __future__ import absolute_import

import itertools
import operator
import textwrap

import packaging.version
import path
import pytest
from six.moves import map

PARDIR = path.Path(__file__).dirname().dirname()

try:
    _accumulate = itertools.accumulate
except AttributeError:

    def _accumulate(iterable, func=operator.add):
        it = iter(iterable)
        try:
            total = next(it)
        except StopIteration:
            return
        yield total
        for element in it:
            total = func(total, element)
            yield total


def last(iterable, default=None):
    it = iter(iterable)
    _last = next(it, default)
    while True:
        try:
            _last = next(it)
        except StopIteration:
            return _last


def accumulate(changelog):
    tags = list(map(operator.itemgetter(1), changelog))
    return last(_accumulate(tags, set.__or__))


@pytest.fixture()
def current_pbr_version(git_command):
    res = git_command("log --decorate=full --format=%h%x00%s%x00%d").splitlines()
    changelog = []
    for line in res:
        parts = line.split("\x00")
        if len(parts) != 3:
            continue

        sha, msg, refname = parts
        tags = set()

        # refname can be:
        #  <empty>
        #  HEAD, tag: refs/tags/1.4.0, refs/remotes/origin/master, \
        #    refs/heads/master
        #  refs/tags/1.3.4
        if "refs/tags/" in refname:
            refname = refname.strip()[1:-1]  # remove wrapping ()'s
            # If we start with "tag: refs/tags/1.2b1, tag: refs/tags/1.2"
            # The first split gives us "['', '1.2b1, tag:', '1.2']"
            # Which is why we do the second split below on the comma
            for tag_string in refname.split("refs/tags/")[1:]:
                # git tag does not allow : or " " in tag names, so we split
                # on ", " which is the separator between elements
                candidate = tag_string.split(", ")[0]
                tags.add(candidate)

        changelog.append((sha, tags, msg))

    tags = accumulate(changelog)
    current = max(tags, key=packaging.version.parse)
    count = itertools.count()
    for sha, tags, msg in changelog:
        step = next(count)
        if current in tags:
            break
    else:
        step = len(changelog)

    v = packaging.version.parse(current)
    if step != 0:
        major, minor, patch = v.release
        patch += 1
        # noinspection PyProtectedMember
        v._version = v._version._replace(
            dev=("post", step), release=(major, minor, patch)
        )

    return str(v)


@pytest.mark.parametrize(
    ("parameter", "expected"),
    [
        ("--name", "format-pipfile"),
        ("--author", "Brandon LeBlanc"),
        ("--author-email", "projects+format-pipfile@leblanc.codes"),
        ("--maintainer", "UNKNOWN"),
        ("--maintainer-email", "UNKNOWN"),
        ("--contact", "Brandon LeBlanc"),
        ("--contact-email", "projects+format-pipfile@leblanc.codes"),
        ("--url", "https://github.com/demosdemon/format-pipfile"),
        ("--license", "MIT"),
        (
            "--description",
            "A Python utility to format TOML Pipfiles with some very opinionated rules.",
        ),
        (
            "--long-description",
            "A Python utility to format TOML Pipfiles with some very opinionated rules.",
        ),
        ("--platforms", "UNKNOWN"),
        (
            "--classifiers",
            """
            Development Status :: 4 - Beta
            Intended Audience :: Developers
            License :: OSI Approved :: MIT License
            Natural Language :: English
            Operating System :: OS Independent
            Programming Language :: Python :: 2
            Programming Language :: Python :: 2.7
            Programming Language :: Python :: 3
            Programming Language :: Python :: 3.4
            Programming Language :: Python :: 3.5
            Programming Language :: Python :: 3.6
            Programming Language :: Python :: 3.7
            Programming Language :: Python :: Implementation :: CPython
            Programming Language :: Python :: Implementation :: PyPy
            Topic :: Software Development
            """,
        ),
        ("--provides", ""),
        ("--keywords", ""),
        ("--requires", ""),
        ("--obsoletes", ""),
    ],
)
def test_metadata(delegator, parameter, expected):
    with PARDIR:
        cmd = "python setup.py {}".format(parameter)
        res = delegator(cmd)

    if "\n" in expected:
        expected = textwrap.dedent(expected).strip()

    assert res == expected


def test_metadata_version(delegator, current_pbr_version):
    with PARDIR:
        cmd = "python setup.py --version"
        res = delegator(cmd)

    assert res == current_pbr_version

    with PARDIR:
        cmd = "python setup.py --fullname"
        res = delegator(cmd)

    assert res == "format-pipfile-{}".format(current_pbr_version)



# -*- coding: utf-8 -*-
"""
AppVeyor will at least have a few Pythons around, so there's no point of implementing a
bootstrapper in PowerShell.

This is a port of
<https://github.com/pypa/python-packaging-user-guide/blob/master/source/code/install.ps1>
with various fixes and improvements that just weren't feasible to implement in
PowerShell.
"""

from __future__ import absolute_import, print_function

from os import environ
from os.path import exists
from shlex import shlex
from subprocess import check_call

from six.moves import range
from six.moves.urllib.request import urlretrieve


def shlex_split(string):
    lex = shlex(string, posix=True)
    lex.whitespace_split = True
    lex.commenters = ""
    return list(lex)


BASE_URL = "https://www.python.org/ftp/python/"
GET_PIP_URL = "https://bootstrap.pypa.io/get-pip.py"
GET_PIP_PATH = "C:\\get-pip.py"

VERSIONS = {
    "2.7": ("2.7.15", "msi"),
    "3.4": ("3.4.4", "msi"),
    "3.5": ("3.5.6", "exe"),
    "3.6": ("3.6.7", "exe"),
    "3.7": ("3.7.1", "exe"),
}
ARITY = {"64": "{0}/python-{0}.amd64.{1}", "32": "{0}/python-{0}.{1}"}

INSTALL_CMD = {
    # Commands are allowed to fail only if they are not the last command.
    # Eg: uninstall (/x) allowed to fail
    "msi": [
        shlex_split("msiexec.exe /L*+! install.log /qn /x {path}"),
        shlex_split("msiexec.exe /L*+! install.log /qn /x {path}"),
    ],
    "exe": [shlex_split("{path} /quiet TargetDir={home}")],
}


def download_file(url, path):
    print(("Downloading: {} (into {})".format(url, path)))
    progress = [0, 0]

    def report(count, size, total):
        progress[0] = count * size
        if progress[0] - progress[1] > 1000000:
            progress[1] = progress[0]
            print(("Downloaded {:,}/{:,} ...".format(progress[1], total)))

    dest, _ = urlretrieve(url, path, reporthook=report)
    return dest


def install_python(version, arch, home):
    print(
        (
            "Installing Python {} for {} bit architecture to {}".format(
                version, arch, home
            )
        )
    )
    if exists(home):
        return

    path, extension = download_python(version, arch)
    print(("Installing {} to {}".format(path, home)))
    success = False
    for cmd in INSTALL_CMD[extension]:
        cmd = [part.format(home=home, path=path) for part in cmd]
        print(("Running: {}".format(" ".join(cmd))))
        try:
            check_call(cmd)
        except Exception as exc:
            print(("Failed command '{}' with: {}".format(" ".join(cmd), exc)))
            if exists("install.log"):
                with open("install.log") as fp:
                    print((fp.read()))
        else:
            success = True

    if success:
        print("Installation complete!")
    else:
        print("Installation failed!")


def download_python(version, arch):
    full_version, extension = VERSIONS[version]
    url = ARITY[arch].format(full_version, extension)
    path = "installer.{}".format(extension)
    for _ in range(3):
        try:
            return download_file(url, path), extension
        except Exception as exc:
            print(("Failed to download: {}".format(exc)))
        print("Retrying...")


def install_pip(home):
    pip_path = home + "/Scripts/pip.exe"
    python_path = home + "/python.exe"
    if exists(pip_path):
        print("pip already installed.")
    else:
        print("Installing pip...")
        download_file(GET_PIP_URL, GET_PIP_PATH)
        print(("Executing: {} {}".format(python_path, GET_PIP_PATH)))
        check_call([python_path, GET_PIP_PATH])


def install_packages(home, *packages):
    cmd = [home + "/Scripts/pip.exe", "install"]
    cmd.extend(packages)
    check_call(cmd)


if __name__ == "__main__":
    install_python(
        environ["PYTHON_VERSION"], environ["PYTHON_ARCH"], environ["PYTHON_HOME"]
    )
    install_pip(environ["PYTHON_HOME"])
    install_packages(
        environ["PYTHON_HOME"],
        "setuptools>=18.0.1",
        "wheel",
        "tox",
        "virtualenv>=13.1.0",
    )



# -*- coding: utf-8 -*-
"""The entry point to the ``format_pipfile`` package."""

from __future__ import absolute_import

if __name__ == "__main__":
    from format_pipfile.cli import main

    main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from __future__ import absolute_import, division, print_function

import os

import jinja2
import matrix
import path


def load_matrix(base_path):
    environments = {}
    items = list(matrix.from_file(base_path / "setup.cfg").items())
    items.insert(0, ("check", {"coverage_flag": "False"}))
    for alias, conf in items:
        environments[alias] = conf.copy()
        pyenv = conf.pop("python_version", "py37")
        windows = not pyenv.startswith("pypy")
        python = pyenv if pyenv.startswith("pypy") else "{0[2]}.{0[3]}".format(pyenv)
        cover = conf.pop("coverage_flag", "true").lower()
        deps = conf.pop("dependencies", "").split()
        env_vars = {k.upper(): v for k, v in list(conf.items())}
        environments[alias].update(
            windows=windows,
            python=python,
            cover=cover == "true",
            deps=deps,
            env_vars=env_vars,
            pyenv=pyenv,
        )

    return environments


def main():
    base_path = path.Path(__file__).dirname().dirname()
    print(("Project path: {}".format(base_path)))

    template_path = base_path / "ci" / "templates"
    jinja = jinja2.Environment(
        loader=jinja2.FileSystemLoader(template_path),
        trim_blocks=True,
        lstrip_blocks=True,
        keep_trailing_newline=False,
        extensions=("jinja2.ext.do",),
    )

    environments = load_matrix(base_path)
    for name in template_path.listdir():
        name = os.path.split(name)[1]
        root = os.path.splitext(name)[0]
        tpl = jinja.get_template(name)
        contents = tpl.render(environments=environments).strip()
        with open(base_path / root, "w") as fp:
            fp.write(contents + "\n")

        print("Wrote {}".format(root))

    print("DONE.")


if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-

from __future__ import absolute_import

from format_pipfile.cli import main


def test_cli_help(cli_runner):
    result = cli_runner.invoke(main, ["--help"], prog_name="format-pipfile")

    assert result.exit_code == 0
    assert "Usage: format-pipfile [OPTIONS]" in result.output
    assert "Update the requirements.txt file and reformat the Pipfile." in result.output
    assert "-r, --requirements-file FILE" in result.output
    assert "--skip-requirements-file" in result.output
    assert "-p, --pipfile FILE" in result.output
    assert "--skip-pipfile" in result.output
    assert "--help" in result.output



# -*- coding: utf-8 -*-

"""
This script performs two tasks:
 * It updates the ``requirements.txt`` file with the packages found in the ``Pipfile``.
 * It reformats the ``Pipfile`` to my opinionated standards.
"""

from __future__ import absolute_import

import difflib
import os
from itertools import filterfalse, starmap
from typing import Any, Callable, Hashable, Optional, Tuple, Union  # noqa

import click
import path
from packaging.utils import canonicalize_name
from requirementslib import Requirement
from tomlkit import loads
from tomlkit.container import Container
from tomlkit.items import AoT, Key, KeyType, Table, Trivia
from tomlkit.toml_document import TOMLDocument  # noqa

MaybeKey = Union[str, Key]
ContainerItem = Tuple[MaybeKey, Any]
SortKeyFunc = Callable[[ContainerItem], Hashable]

ROOT_PATH = os.getcwd()
ROOT_REQUIREMENTS = os.path.join(ROOT_PATH, "requirements.txt")
ROOT_PIPFILE = os.path.join(ROOT_PATH, "Pipfile")
PIPFILE_SECTIONS = ("source", "packages", "dev-packages", "requires", "scripts")
DEFAULT_SOURCE = (
    ("name", "pypi"),
    ("url", "https://pypi.org/simple"),
    ("verify_ssl", True),
)


def unwrap_key(key):
    # type: (MaybeKey) -> str
    if isinstance(key, Key):
        return key.key

    return key


def reorder_container(container, sort_key=None, reverse=False, key_type=None):
    # type: (Container, SortKeyFunc, bool, KeyType) -> None
    items = container.body
    if sort_key is not None or reverse:
        items.sort(key=sort_key, reverse=reverse)

    container._map = {}
    container._body = []

    for key, value in items:
        if key is None:
            pass
        elif isinstance(key, Key):
            if key_type is not None:
                key = Key(key.key, key_type)
        else:
            key = Key(str(key), key_type)

        container.append(key, value)


def pipfile_source_key(pair):
    # type: (ContainerItem) -> Tuple[int, Optional[str]]
    not_found = len(DEFAULT_SOURCE)

    key, value = pair
    if key is None:
        return (not_found + 1, key)

    key = unwrap_key(key)
    norm = key.lower()
    for idx, (key, _) in enumerate(DEFAULT_SOURCE):
        if norm == key:
            return (idx, key)

    return (not_found, key)


def pipfile_section_key(pair):
    # type: (ContainerItem) -> Tuple[int, Optional[str], Optional[str]]
    not_found = len(PIPFILE_SECTIONS)

    key, value = pair
    if key is None:
        return (not_found + 1, None, None)

    key = unwrap_key(key)
    normalized = key.lower().replace("_", "-")

    try:
        idx = PIPFILE_SECTIONS.index(normalized)
    except ValueError:
        idx = not_found

    return (idx, normalized, key)


def pipfile_packages_key(pair):
    # type: (ContainerItem) -> Tuple[bool, Optional[str]]
    key, value = pair
    key = unwrap_key(key)
    return (key is None, canonicalize_name(key) if key else None)


def requirement_sort_key(item):
    # type: (Requirement) -> Tuple[bool, Optional[str]]
    return (item.is_named, item.normalized_name)


def is_default(source):
    # type: (Table) -> bool
    return all(source.get(k) == v for (k, v) in DEFAULT_SOURCE)


def load_pipfile(p=None):
    # type: (Optional[path.Path]) -> TOMLDocument
    if p is None:
        p = path.Path(ROOT_PIPFILE)

    with open(p, "r") as fp:
        doc = loads(fp.read())
    doc._parsed = True

    sources = []
    have_default = False
    for key in ("source", "sources"):
        try:
            item = doc.item(key)  # type: Union[AoT, Table]
        except KeyError:
            continue

        doc.remove(key)

        if isinstance(item, AoT):
            items = item.body
        else:
            items = item.value

        for source in items:
            if not isinstance(source, Table):
                source = Table(source, Trivia(trail=""), is_aot_element=True)
            container = source.value
            reorder_container(container, pipfile_source_key, key_type=KeyType.Basic)
            if not have_default and is_default(container):
                have_default = True

            sources.append(source)

    if not have_default:
        source = Table(Container(True), Trivia(), True)
        for k, v in DEFAULT_SOURCE:
            source.append(k, v)

        sources.insert(0, source)

    doc.append("source", AoT(sources, parsed=True))
    return doc


def update_requirements(p, pipfile):
    # type: (path.Path, TOMLDocument) -> None
    sources = list(filterfalse(is_default, pipfile.get("source", [])))
    packages = pipfile.get("packages", {})
    requirements = list(starmap(Requirement.from_pipfile, list(packages.items())))
    requirements.sort(key=requirement_sort_key)
    lines = [r.as_line(sources) + os.linesep for r in requirements]

    with open(p, "w") as fp:
        fp.writelines(lines)


def dump_pipfile(p, pipfile):
    # type: (path.Path, TOMLDocument) -> None
    reorder_container(pipfile, pipfile_section_key, key_type=KeyType.Bare)

    for name in ("packages", "dev-packages"):
        try:
            section = pipfile.item(name)  # type: Table
        except KeyError:
            pass
        else:
            container = section.value
            reorder_container(container, pipfile_packages_key, key_type=KeyType.Basic)

    value = pipfile.as_string()
    try:
        with open(p, "r") as fp:
            old_value = fp.read()
    except OSError:
        old_value = ""

    if value != old_value:
        diff = difflib.ndiff(old_value.splitlines(True), value.splitlines(True))
        click.echo("".join(diff), nl=False)

        with open(p, "w") as fp:
            fp.write(value)


@click.command()
@click.option(
    "-r",
    "--requirements-file",
    default=ROOT_REQUIREMENTS,
    help="Specify the requirements.txt file to update.",
    show_default=True,
    show_envvar=True,
    type=click.Path(dir_okay=False, readable=True),
)
@click.option(
    "--skip-requirements-file",
    help="Skip updating the requirements file.",
    is_flag=True,
    show_envvar=True,
)
@click.option(
    "-p",
    "--pipfile",
    default=ROOT_PIPFILE,
    help="Specify the Pipfile to read and optionally update.",
    show_default=True,
    show_envvar=True,
    type=click.Path(exists=True, dir_okay=False, readable=True),
)
@click.option(
    "--skip-pipfile",
    is_flag=True,
    help="Skip reformatting the Pipfile.",
    show_envvar=True,
)
def main(requirements_file, skip_requirements_file, pipfile, skip_pipfile):
    # type: (str, bool, str, bool) -> None
    """Update the requirements.txt file and reformat the Pipfile."""
    pipfile_path = path.Path(pipfile)
    pf = load_pipfile(pipfile_path)

    if not skip_requirements_file:
        requirements_file_path = path.Path(requirements_file)
        update_requirements(requirements_file_path, pf)

    if not skip_pipfile:
        dump_pipfile(pipfile_path, pf)


if __name__ == "__main__":
    main()
